@@ -33,14 +33,16 @@
 #define _ABT_H
 /* An array based binary tree that stores ints */
 
+#include "arpcache.h"
+
 typedef struct _abt {
   int size;
   int space;
-  int *data;
+  arp_entry *data;
 } abt;
 
 abt *new_abt();
-void add_abt(abt *, int);
-int  chk_abt(abt *, int);
+void add_abt(abt *, arp_entry);
+int  chk_abt(abt *, arp_entry);
 
 #endif
