@@ -18,7 +18,7 @@
  * The Creators of Wackamole are:
  *  Yair Amir, Ryan Caudy, Aashima Munjal and Theo Schlossnagle.
  *
- *  Copyright (C) 2000-2001 The Johns Hopkins University
+ *  Copyright (C) 2000-2004 The Johns Hopkins University
  *  All Rights Reserved.
  *
  *  This product uses the Spread toolkit, developed by Spread Concepts LLC.
@@ -41,6 +41,10 @@
 #include "arpcache.h"
 #include "spoofmanager.h"
 #include "control.h"
+#include "userloader.h"
+#ifdef USE_EMBEDDED_PERL
+#include "perl.h"
+#endif
 
 /* message types */
 #define		STATE_MESS	1
@@ -111,7 +115,7 @@ static	void	Handle_state();
 static	void	Handle_mature();
 static	void	Handle_balance();
 static	void	Handle_operator();
-static	void	Handle_arp_cache(int *, int);
+static	void	Handle_arp_cache(arp_entry *, int);
 
 static	void	Send_state_message();
 static  void	Send_local_arp_cache_repeat();
@@ -126,8 +130,10 @@ static  void    Claim_unclaimed();
 static  entry  *findVEFromAddress( address Addr );
 /* static  void    AcquireAddress( address Addr ); */
 static  void    Acquire( entry *VE );
+static  void    Handle_Post_Acquire();
 static  void    ReleaseAddress( address Addr );
 static  void    Release( entry *VE );
+static  void    Handle_Post_Release();
 static  void    Balance();
 static  void    Print_alloc();
 /* static  void    String_addr( address Addr, char IP[16] ); */
@@ -136,66 +142,65 @@ static  void    Sig_handler(int signum);
 	void    Spread_reconnect(int ret);
 
 /*************************************************/
-
+static void notice() {
+  wack_alarm( PRINT, "\n/==============================================================================\\");
+  wack_alarm( PRINT, "| The Wackamole Program.                                                       |");
+  wack_alarm( PRINT, "| Copyright (c) 2000-2004 The Johns Hopkins University                         |"); 
+  wack_alarm( PRINT, "| All rights reserved.                                                         |");
+  wack_alarm( PRINT, "|                                                                              |");
+  wack_alarm( PRINT, "| Wackamole is developed at the Center for Networking and Distributed Systems, |");
+  wack_alarm( PRINT, "| The Johns Hopkins University.                                                |");
+  wack_alarm( PRINT, "|                                                                              |");
+  wack_alarm( PRINT, "| The Wackamole package is licensed under the CNDS Open Source License         |");
+  wack_alarm( PRINT, "| You may only use this software in compliance with the License.               |");
+  wack_alarm( PRINT, "| A copy of the license can be found at                                        |");
+  wack_alarm( PRINT, "| http://www.backhand.org/wackamole/license                                    |");
+  wack_alarm( PRINT, "|                                                                              |");
+  wack_alarm( PRINT, "| This product uses the Spread toolkit, developed by Spread Concepts LLC.      |");
+  wack_alarm( PRINT, "| For more information about Spread see http://www.spread.org                  |");
+  wack_alarm( PRINT, "|                                                                              |");
+  wack_alarm( PRINT, "| This software is distributed on an \"AS IS\" basis, WITHOUT WARRANTY OF        |");
+  wack_alarm( PRINT, "| ANY KIND, either express or implied.                                         |");
+  wack_alarm( PRINT, "|                                                                              |");
+  wack_alarm( PRINT, "| Creators:                                                                    |");
+  wack_alarm( PRINT, "|    Yair Amir             yairamir@cnds.jhu.edu                               |");
+  wack_alarm( PRINT, "|    Ryan Caudy            wyvern@cnds.jhu.edu                                 |");
+  wack_alarm( PRINT, "|    Aashima Munjal        munjal@jhu.edu                                      |");
+  wack_alarm( PRINT, "|    Theo Schlossnagle     jesus@cnds.jhu.edu                                  |");
+  wack_alarm( PRINT, "|                                                                              |");
+  wack_alarm( PRINT, "| For a full list of contributors, see Readme.txt in the distribution.         |");
+  wack_alarm( PRINT, "|                                                                              |");
+  wack_alarm( PRINT, "| WWW:     www.backhand.org     www.cnds.jhu.edu                               |");
+  wack_alarm( PRINT, "| Contact: wackamole@backhand.org                                              |");
+  wack_alarm( PRINT, "|                                                                              |");
+  wack_alarm( PRINT, "| Version %-20s                                                 |", VERSION_STRING);
+  wack_alarm( PRINT, "| Released %-20s                                                |", VERSION_RELEASE_DATE); 
+  wack_alarm( PRINT, "|                                                                              |");
+  wack_alarm( PRINT, "\\==============================================================================|");
+}
 int main( int argc, char *argv[] )
 {
   int		fd;
   struct sigaction signalaction;
  
   E_init();
-  Alarm_set( PRINT | EXIT ); 
+  wack_alarm_set( PRINT | EXIT ); 
  
   Usage( argc, argv );
   Wackamole_init();
 
-  Alarm( PRINT, "\n/==============================================================================\\");
-  Alarm( PRINT, "| The Wackamole Program.                                                       |");
-  Alarm( PRINT, "| Copyright (c) 2000-2001 The Johns Hopkins University                         |"); 
-  Alarm( PRINT, "| All rights reserved.                                                         |");
-  Alarm( PRINT, "|                                                                              |");
-  Alarm( PRINT, "| Wackamole is developed at the Center for Networking and Distributed Systems, |");
-  Alarm( PRINT, "| The Johns Hopkins University.                                                |");
-  Alarm( PRINT, "|                                                                              |");
-  Alarm( PRINT, "| The Wackamole package is licensed under the CNDS Open Source License         |");
-  Alarm( PRINT, "| You may only use this software in compliance with the License.               |");
-  Alarm( PRINT, "| A copy of the license can be found at                                        |");
-  Alarm( PRINT, "| http://www.backhand.org/wackamole/license                                    |");
-  Alarm( PRINT, "|                                                                              |");
-  Alarm( PRINT, "| This product uses the Spread toolkit, developed by Spread Concepts LLC.      |");
-  Alarm( PRINT, "| For more information about Spread see http://www.spread.org                  |");
-  Alarm( PRINT, "|                                                                              |");
-  Alarm( PRINT, "| This software is distributed on an \"AS IS\" basis, WITHOUT WARRANTY OF        |");
-  Alarm( PRINT, "| ANY KIND, either express or implied.                                         |");
-  Alarm( PRINT, "|                                                                              |");
-  Alarm( PRINT, "| Creators:                                                                    |");
-  Alarm( PRINT, "|    Yair Amir             yairamir@cnds.jhu.edu                               |");
-  Alarm( PRINT, "|    Ryan Caudy            wyvern@cnds.jhu.edu                                 |");
-  Alarm( PRINT, "|    Aashima Munjal        munjal@jhu.edu                                      |");
-  Alarm( PRINT, "|    Theo Schlossnagle     jesus@cnds.jhu.edu                                  |");
-  Alarm( PRINT, "|                                                                              |");
-  Alarm( PRINT, "| For a full list of contributors, see Readme.txt in the distribution.         |");
-  Alarm( PRINT, "|                                                                              |");
-  Alarm( PRINT, "| WWW:     www.backhand.org     www.cnds.jhu.edu                               |");
-  Alarm( PRINT, "| Contact: wackamole@backhand.org                                              |");
-  Alarm( PRINT, "|                                                                              |");
-  Alarm( PRINT, "| Version %-7s Released %-20s                                |", VERSION_STRING, VERSION_RELEASE_DATE); 
-  Alarm( PRINT, "|                                                                              |");
-  Alarm( PRINT, "\\==============================================================================|");
-
-  if(Debug) 
-    Alarm_set( PRINT | ARPING | DEBUG | EXIT );
+  if(Debug) wack_alarm_set( PRINT | ARPING | WACK_DEBUG | EXIT );
   else {
-
     char pidstring[10];
 
-    Alarm_enable_syslog("wackamole");
-    Alarm_set(PRINT | EXIT);
+    wack_alarm_set(PRINT | EXIT);
+    wack_alarm_enable_syslog("wackamole");
 
     if(fork()) exit(1);
     setsid();
     if(fork()) exit(1);
     if( chdir("/") != 0 ){
-      Alarm(PRINT,"chdir to root failed");
+      wack_alarm(PRINT,"chdir to root failed");
     }
 
     umask(0027);
@@ -204,7 +209,7 @@ int main( int argc, char *argv[] )
     fd = open(_PATH_WACKAMOLE_PIDDIR "/wackamole.pid",
 	      O_WRONLY|O_CREAT|O_TRUNC, 0644);
     if(fd < 0) {
-      Alarm(EXIT,
+      wack_alarm(EXIT,
 	    "Cannot write PID file " _PATH_WACKAMOLE_PIDDIR "/wackamole.pid");
     }
     snprintf(pidstring, 10, "%d\n", pid);
@@ -217,9 +222,17 @@ int main( int argc, char *argv[] )
 
     write(fd, pidstring, strlen(pidstring));
     close(fd);
+    fd = open("/dev/null", O_RDWR);
+    if(fd <= 2) {
+      wack_alarm(EXIT, "Cannot open /dev/null\n");
+    }
     close(STDIN_FILENO);
     close(STDOUT_FILENO);
     close(STDERR_FILENO);
+    dup2(fd,STDIN_FILENO);
+    dup2(fd,STDOUT_FILENO);
+    dup2(fd,STDERR_FILENO);
+    close(fd);
   }
 
   signalaction.sa_handler = Sig_handler;
@@ -228,22 +241,22 @@ int main( int argc, char *argv[] )
 
 
   if(sigaction(SIGINT, &signalaction, NULL)) {
-    Alarm(EXIT, "An error occured while registering a SIGINT handler");
+    wack_alarm(EXIT, "An error occured while registering a SIGINT handler");
   }
 
   if(sigaction(SIGTERM, &signalaction, NULL)) {
-    Alarm(EXIT, "An error occured while registering a SIGTERM handler");
+    wack_alarm(EXIT, "An error occured while registering a SIGTERM handler");
   }
   if(sigaction(SIGBUS, &signalaction, NULL)) {
-    Alarm(EXIT, "An error occured while registering a SIGBUS handler");
+    wack_alarm(EXIT, "An error occured while registering a SIGBUS handler");
   }
 
   if(sigaction(SIGQUIT, &signalaction, NULL)) {
-    Alarm(EXIT, "An error occured while registering a SIGQUIT handler");
+    wack_alarm(EXIT, "An error occured while registering a SIGQUIT handler");
   }
 
   if(sigaction(SIGSEGV, &signalaction, NULL)) {
-    Alarm(EXIT, "An error occured while registering a SIGSEGV handler");
+    wack_alarm(EXIT, "An error occured while registering a SIGSEGV handler");
   }
 
   if_initialize();
@@ -287,12 +300,12 @@ void	Handle_network()
       else if	( Mess_type == MATURING_MESS ) Handle_mature();
       else if	( Mess_type == BALANCE_MESS  ) Handle_balance();
       else if	( Mess_type == OPERATOR_MESS ) Handle_operator();
-      else if	( Mess_type == ARPCACHE_MESS ) Handle_arp_cache((int *)Mess,ret);
+      else if	( Mess_type == ARPCACHE_MESS ) Handle_arp_cache((arp_entry *)Mess,ret);
       
       /* Ignore messages of other types */
       
     }else{
-      Alarm(EXIT,"Error: received message of unknown message type %d with ret %d",
+      wack_alarm(EXIT,"Error: received message of unknown message type %d with ret %d",
 	     Service_type, ret );
      }
   
@@ -311,7 +324,7 @@ static	void	Turn_mature()
       Spread_reconnect(ret);
     }
   
-  Alarm(LOGIC, "Turn_mature");
+  wack_alarm(LOGIC, "Turn_mature");
 }
 
 /*************************************************/
@@ -323,13 +336,13 @@ static	void	Handle_mature()
   Maturity = 1;
   E_dequeue( Turn_mature, 0, 0 );
   
-  Alarm(LOGIC,"Handle_mature");
+  wack_alarm(LOGIC,"Handle_mature");
   if ( State != GATHER )
     {
       /* ### deterministically grab ip addresses */
       Priority_claim();
       Claim_unclaimed();
-      Alarm(LOGIC, "Shifting to RUN in Handle_mature()" );
+      wack_alarm(LOGIC, "Shifting to RUN in Handle_mature()" );
       State = RUN;
       E_queue( Balance, 0, 0, Balance_timer );
     }
@@ -339,18 +352,19 @@ static	void	Handle_mature()
 
 static	void	Handle_operator()
 {
-  Alarm(LOGIC,"Handle_operator");
+  wack_alarm(LOGIC,"Handle_operator");
   
 }
 
 /*************************************************/
 
-static	void	Handle_arp_cache(int *addresses, int size)
+static	void	Handle_arp_cache(arp_entry *arps, int size)
 {
   int i;
-  size /= sizeof(int);
+  size /= sizeof(arp_entry);
+  wack_alarm(WACK_DEBUG, "Reading %d shared arp entries", size);
   for(i=0;i<size;i++)
-    insert_arp_cache_shared(ntohl(addresses[i]));
+    insert_arp_cache_shared(arps[i]);
 }
 
 /*************************************************/
@@ -360,10 +374,12 @@ static	void	Handle_balance()
   int        i, j, k;
   int        num_bytes;
   int        *num_balanced;
+  int	     num_released;
+  int	     num_acquired;
   address    *curr_real_ptr;
   address    *curr_pseudo_ptr;
 
-  Alarm(LOGIC,"Handle_balance");
+  wack_alarm(LOGIC,"Handle_balance");
   if( State != RUN ) return;
   Print_alloc();
 
@@ -372,6 +388,8 @@ static	void	Handle_balance()
   num_bytes += sizeof( int );
   if( Endian_mismatch ) *num_balanced = Flip_int32( *num_balanced );
 
+  num_released = 0;
+  num_acquired = 0;
   for ( k = 0; k < *num_balanced; k++ )
     {
       curr_real_ptr = (address *)&Mess[num_bytes];
@@ -389,10 +407,14 @@ static	void	Handle_balance()
 	{
 	  if( *curr_pseudo_ptr == _pif_ip_s(Allocation_table[i]) )
 	    {
-	      if ( _rif_ip_s(My) == *curr_real_ptr )
+	      if ( _rif_ip_s(My) == *curr_real_ptr ) {
+                num_acquired++;
 		Acquire( &Allocation_table[i] );
-	      if ( _rif_ip_s(My) == _rif_ip_s(Allocation_table[i]) )
+              }
+	      if ( _rif_ip_s(My) == _rif_ip_s(Allocation_table[i]) ) {
+                num_released++;
 		Release( &Allocation_table[i] );
+              }
 	      for ( j = 0; j < Num_members; j++ )
 		{
 		  if ( _rif_ip_s(Members[j]) == _rif_ip_s(Allocation_table[i]) )
@@ -405,8 +427,13 @@ static	void	Handle_balance()
 	    }  
 	}
     }
-  
-  Alarm(LOGIC,  "Finished Handle_balance." );
+  if(num_acquired) {
+    Handle_Post_Acquire();
+  }
+  if(num_released) {
+    Handle_Post_Release();
+  }
+  wack_alarm(LOGIC,  "Finished Handle_balance." );
   Print_alloc();
 }
 
@@ -416,13 +443,13 @@ static	void	Handle_membership()
 {
   int	i;
   
-  Alarm(LOGIC, "Handle_membership");
+  wack_alarm(LOGIC, "Handle_membership");
   
   if( strcmp( Sender, Spread_group ) != 0 )
     {
       Bye();
       Clean_up();
-      Alarm(EXIT, "Handle_membership:  Bug! got a membership for group %s",
+      wack_alarm(EXIT, "Handle_membership:  Bug! got a membership for group %s",
 	     Sender );
       
     }
@@ -435,12 +462,6 @@ static	void	Handle_membership()
   for( i = 0; i < Num_members; i++ )
     {
       memcpy( Members[i].private_group_name, Target_groups[i], MAX_GROUP_NAME );
-      /* RWC: SPREAD GIVES US THIS INFO, WHY WORK FOR IT? */
-#if 0
-      if( strcmp( Target_groups[i], Private_group ) == 0 ){
-	My_index = i;
-      }
-#endif
       My_index = Mess_type;
       _rif_ip_s(Members[i]) = 0;
       Members[i].num_allocated = 0;
@@ -455,9 +476,9 @@ static	void	Handle_membership()
       Allocation_table[i].claim_priority = UNCLAIMED;
     }
   Send_state_message();		/* ### */
-  Alarm(LOGIC,  "Shifting to GATHER" );
+  wack_alarm(LOGIC,  "Shifting to GATHER" );
   State = GATHER;
-  Alarm(LOGIC,  "Dequeuing Balance" ); /* Debug */
+  wack_alarm(LOGIC,  "Dequeuing Balance" ); /* Debug */
   E_dequeue( Balance, 0, 0 );
 
 }
@@ -469,6 +490,7 @@ static	void	Handle_state()
   group_id	*curr_gid_ptr;
   address	*curr_real_address_ptr;
   int		num_bytes;
+  int		num_released;
   int           *curr_index_ptr;
   int           *curr_maturity_ptr;
   int           *curr_num_allocated_ptr;  
@@ -478,7 +500,7 @@ static	void	Handle_state()
   address       *curr_prefer_address;
   int           i, j, k;
 
-  Alarm(LOGIC,  "Handle state" );
+  wack_alarm(LOGIC,  "Handle state" );
 
   num_bytes = 0;
   curr_gid_ptr = (group_id *)&Mess[num_bytes];
@@ -516,6 +538,7 @@ static	void	Handle_state()
   Members[*curr_index_ptr].num_allocated = *curr_num_allocated_ptr;
   Members[*curr_index_ptr].got_state_from = 1;
 
+  num_released = 0;
   /* Rip addresses and compute stuff */
   for ( i = 0; i < *curr_num_allocated_ptr; i++ )
     {
@@ -545,6 +568,7 @@ static	void	Handle_state()
 		      Members[*curr_index_ptr].num_allocated--;
 		      if( _rif_ip_s(My) == *curr_real_address_ptr )
 			{
+                          num_released++;
 			  ReleaseAddress(*curr_pseudo_addr_ptr);
 			}
 		      Old_table[j].claim_priority = Allocation_table[j].claim_priority;
@@ -554,7 +578,10 @@ static	void	Handle_state()
 			   *curr_claim_priority_ptr > PREFER )
 		    {
 		      if( _rif_ip_s(My) == _rif_ip_s(Allocation_table[j]) )
-			ReleaseAddress( *curr_pseudo_addr_ptr );
+                        {
+                          num_released++;
+			  ReleaseAddress( *curr_pseudo_addr_ptr );
+                        }
 		      for(k = 0; k < Num_members; ++k)
 			{
 			  if( _rif_ip_s(Members[k]) == _rif_ip_s(Allocation_table[j]) )
@@ -570,7 +597,10 @@ static	void	Handle_state()
 	    }
 	}
     }
-  
+ 
+  if(num_released) {
+    Handle_Post_Release();
+  } 
   curr_num_prefer_ptr = (int *)&Mess[num_bytes];
   num_bytes += sizeof( int );
   if( Endian_mismatch ) *curr_num_prefer_ptr = Flip_int32( *curr_num_prefer_ptr );
@@ -597,11 +627,11 @@ static	void	Handle_state()
   if ( Maturity == 0 && *curr_maturity_ptr == 1 )
     Handle_mature();
 				   
-  Alarm(LOGIC, "handle state:  got state from %d", *curr_real_address_ptr );
+  wack_alarm(LOGIC, "handle state:  got state from %d", *curr_real_address_ptr );
   Num_pending_states--;
   if( Num_pending_states == 0 )
     {
-      Alarm(LOGIC,  "handle state: Finished getting states, maturity is %d.", Maturity );
+      wack_alarm(LOGIC,  "handle state: Finished getting states, maturity is %d.", Maturity );
       if ( Maturity ) 
 	{
 	  /* wackamole calculate, drop and aquire */
@@ -611,12 +641,12 @@ static	void	Handle_state()
 
 	  State = RUN;
 	  E_queue( Balance, 0, 0, Balance_timer );
-	  Alarm(LOGIC,  "handle state: shifting to RUN");
+	  wack_alarm(LOGIC,  "handle state: shifting to RUN");
 	}
       else
 	{
 	  State = BOOT;
-	  Alarm(LOGIC,  "handle state: shifting back to BOOT");
+	  wack_alarm(LOGIC,  "handle state: shifting back to BOOT");
 	}
       Print_alloc();
     }
@@ -643,7 +673,7 @@ static	void	Send_state_message()
    *            pseudo
    */
 
-  Alarm(LOGIC, "sending state");/*debug*/
+  wack_alarm(LOGIC, "sending state");/*debug*/
   Print_alloc();
   num_bytes = 0;
   memcpy( &Mess[num_bytes], &Gid, sizeof( group_id ) );
@@ -697,18 +727,18 @@ static	void	Send_local_arp_cache_repeat()
 static	void	Send_local_arp_cache()
 {
   int ret, num_bytes=0;
-  int *s, *addresses;
-  int *d = (int *)Mess;
+  arp_entry *addresses, *s;
+  arp_entry *d = (arp_entry *)Mess;
 
   sample_arp_cache();
   s = addresses = reference_private_arp_cache();
 
-  while(*s && (char *)d < &Mess[MAX_MESS_LEN-sizeof(int)]) {
-    *d = htonl(*s);
+  while(s->ip && (char *)d < &Mess[MAX_MESS_LEN-sizeof(arp_entry)]) {
+    memcpy(d, s, sizeof(arp_entry));
     d++; s++;
-    num_bytes += sizeof(int);
+    num_bytes += sizeof(arp_entry);
   }
-  Alarm(DEBUG, "Sending %d local arp entries", num_bytes/sizeof(int));
+  wack_alarm(WACK_DEBUG, "Sending %d local arp entries", num_bytes/sizeof(arp_entry));
   ret = SP_multicast( Mbox, RELIABLE_MESS, Spread_group, ARPCACHE_MESS, num_bytes, Mess );  
   if( ret < 0 )
     {
@@ -722,17 +752,26 @@ static	void	Send_local_arp_cache()
 static  void    Priority_claim()
 {
   int        i, j;
-  Alarm(LOGIC,  "Priority_claim called" );
+  int        num_acquired;
+  int        num_released;
+
+  num_acquired = 0;
+  num_released = 0;
+  wack_alarm(LOGIC,  "Priority_claim called" );
   for ( i = 0; i < Num_pseudo; i++ )
     {
       if ( Allocation_table[i].claim_priority == PREFER )
 	{
 	  if ( _rif_ip_s(My) == _rif_ip_s(Allocation_table[i]) &&
-	       _rif_ip_s(My) != _rif_ip_s(Old_table[i]) )
+	       _rif_ip_s(My) != _rif_ip_s(Old_table[i]) ) {
+            num_acquired++;
 	    Acquire( &Allocation_table[i] );
+          }
 	  if ( _rif_ip_s(My) == _rif_ip_s(Old_table[i]) &&
-	       _rif_ip_s(My) != _rif_ip_s(Allocation_table[i]) )
+	       _rif_ip_s(My) != _rif_ip_s(Allocation_table[i]) ) {
+            num_released++;
 	    Release( &Allocation_table[i] );
+          }
 	  for ( j = 0; j < Num_members; j++ )
 	    {
 	      if ( _rif_ip_s(Members[j]) == _rif_ip_s(Old_table[i]) &&
@@ -749,7 +788,10 @@ static  void    Priority_claim()
 	  Allocation_table[i].claim_priority = PREFER_CLAIMED;
 	}
     }
-  
+
+  if(num_released) {
+    Handle_Post_Release();  
+  }
   /** If we may need to revalidate num_allocated data. ***/
   if ( Maturity && !Old_maturity )
     {
@@ -770,15 +812,11 @@ static  void    Claim_unclaimed()
 {
   int         i, j;
   int         Member_to_claim;
-  Alarm(LOGIC,  "Claim_unclaimed called" );
+  wack_alarm(LOGIC,  "Claim_unclaimed called" );
   for ( i = 0; i < Num_pseudo; i++ )
     if ( Allocation_table[i].claim_priority == UNCLAIMED )
       {
 	Member_to_claim = 0;
-        /* RWC: START COMPARING WITH INDEX 1, NOT INDEX 0 */
-#if 0
-        for ( j = 0; j < Num_members; j++ )
-#endif
         for ( j = 1; j < Num_members; j++ )
 	  if ( Members[j].num_allocated < Members[Member_to_claim].num_allocated )
             Member_to_claim = j;
@@ -806,11 +844,41 @@ static  entry  *findVEFromAddress( address Addr )
   struct in_addr ipaddr;
   ipaddr.s_addr = Addr;
   if(!VE) {
-    Alarm(PRINT, "AcquireAddress: Can't find VirtualInterface for: %s", inet_ntoa(ipaddr));
+    wack_alarm(PRINT, "AcquireAddress: Can't find VirtualInterface for: %s", inet_ntoa(ipaddr));
   } else {
     Acquire(VE);
   }
 } */
+#ifdef NEEDS_FORCE_REUP
+static  void    force_reup()
+{
+  int i;
+  for(i=0; i < Num_pseudo; i++) {
+    if ( _rif_ip_s(My) == _rif_ip_s(Allocation_table[i]) ) {
+      /* This is ours */
+      int ic, n=0;
+      struct interface iface, *nif;
+      entry *VE;
+
+      VE = &Allocation_table[i];
+      for(nif=&VE->pseudo_if; n<MAX_DEP_IF; nif = &(VE->extra_ifs[n++])) {
+	if(nif->ipaddr.s_addr == 0) break;
+        memcpy(&iface, nif, sizeof(struct interface));
+        ic = if_up(&iface);
+        if(ic)
+          wack_alarm(PRINT, "%s", if_error());
+        else {
+          char buffer[16];
+          snprintf(buffer, 16, inet_ntoa(iface.ipaddr));
+          wack_alarm(PRINT, "  (re)UP: %s:%s/%s",
+                     iface.ifname,buffer,inet_ntoa(iface.netmask));
+        }
+      }
+      invoke_spoofer(VE);
+    }
+  }
+}
+#endif
 static  void    Acquire( entry *VE )
 {
   int ic, n=0;
@@ -819,23 +887,25 @@ static  void    Acquire( entry *VE )
   for(nif=&VE->pseudo_if; n<MAX_DEP_IF; nif = &(VE->extra_ifs[n++])) {
 	if(nif->ipaddr.s_addr == 0) break;
     memcpy(&iface, nif, sizeof(struct interface));
-    iface.ipaddr.s_addr = htonl(iface.ipaddr.s_addr);
-    iface.netmask.s_addr = htonl(iface.netmask.s_addr);
-    iface.bcast.s_addr = htonl(iface.bcast.s_addr);
     ic = if_up(&iface);
     if(ic)
-      Alarm(PRINT, "%s", if_error());
+      wack_alarm(PRINT, "%d %s", __LINE__, if_error());
     else {
       char buffer[16];
       snprintf(buffer, 16, inet_ntoa(iface.ipaddr));
-      Alarm(PRINT, "  UP: %s:%s/%s",
+      wack_alarm(PRINT, "  UP: %s:%s/%s",
 		iface.ifname,buffer,inet_ntoa(iface.netmask));
     }
   }
   invoke_spoofer(VE);
+  execute_all_user(VE->pseudo_if, VE->extra_ifs, VE->real_if,
+                   DLFUNCS_TYPE_ON_UP);
   My.num_allocated++;
 }
 
+static void Handle_Post_Acquire() {
+  execute_all_user_simple(DLFUNCS_TYPE_POST_UP);
+}
 /***************************************************/
 
 static  void    ReleaseAddress( address Addr )
@@ -844,7 +914,7 @@ static  void    ReleaseAddress( address Addr )
   struct in_addr ipaddr;
   ipaddr.s_addr = Addr;
   if(!VE) {
-    Alarm(PRINT, "ReleaseAddress: Can't find VirtualInterface for: %s", inet_ntoa(ipaddr));
+    wack_alarm(PRINT, "ReleaseAddress: Can't find VirtualInterface for: %s", inet_ntoa(ipaddr));
   } else {
     Release(VE);
   }
@@ -857,39 +927,50 @@ static  void    Release( entry *VE )
   for(nif=&VE->pseudo_if; n<MAX_DEP_IF; nif = &(VE->extra_ifs[n++])) {
 	if(nif->ipaddr.s_addr == 0) break;
     memcpy(&idown, nif, sizeof(struct interface));
-    idown.ipaddr.s_addr = htonl(idown.ipaddr.s_addr);
-    idown.netmask.s_addr = htonl(idown.netmask.s_addr);
-    idown.bcast.s_addr = htonl(idown.bcast.s_addr);
     ic = if_down(&idown);
     if(ic) {
       const char *em = if_error();
       if(em && strlen(em)) {
-      	Alarm(PRINT, "%s", if_error());
+      	wack_alarm(PRINT, "%d %s", __LINE__, if_error());
       }
     } else {
       char buffer[16];
       snprintf(buffer, 16, inet_ntoa(idown.ipaddr));
-      Alarm(PRINT, "DOWN: %s:%s/%s",
+      wack_alarm(PRINT, "DOWN: %s:%s/%s",
 	idown.ifname,buffer,inet_ntoa(idown.netmask));
     }
+    execute_all_user(VE->pseudo_if, VE->extra_ifs, VE->real_if,
+                     DLFUNCS_TYPE_ON_DOWN);
   }
   My.num_allocated--;
 }
 
+static void Handle_Post_Release() {
+#ifdef NEEDS_FORCE_REUP
+  force_reup();
+#endif
+  execute_all_user_simple(DLFUNCS_TYPE_POST_DOWN);
+}
 
 /***************************************************/
 
 static  void    Balance()
 {
-  int          num_bytes;
-  int          ret;
-  int          i, j, k;
-  int          min_index;
-  int          max_index;
-  member       tempMembers[MAX_PSEUDO];
-  entry        tempTable[MAX_PSEUDO];
-  Alarm(LOGIC, "Balance called." );
+  int            ret;
+  int            num_bytes;
+  int            i, j, k;
+  int            min_index;
+  int            max_index;
+  static member *tempMembers = NULL;
+  static entry  *tempTable = NULL;
+  wack_alarm(LOGIC, "Balance called." );
 
+  if(!tempMembers) {
+    tempMembers = (member *)malloc(sizeof(member) * MAX_PSEUDO);
+  }
+  if(!tempTable) {
+    tempTable = (entry *)malloc(sizeof(entry) * MAX_PSEUDO);
+  }
   if( My_index ) goto leave;
 
   if( State != RUN ) return;
@@ -962,7 +1043,7 @@ static  void    Balance()
       Spread_reconnect(ret);
     }
   State = RUN;
-  Alarm(LOGIC, "Shifting to RUN in balance" );
+  wack_alarm(LOGIC, "Shifting to RUN in balance" );
  leave:
   if( !Complete_balance )
     E_queue( Balance, 0, 0, Balance_timer );
@@ -973,7 +1054,7 @@ static  void    Balance()
 
 static	void    Wackamole_init()
 {
-  Alarm(LOGIC,"Wackamole_init");
+  wack_alarm(LOGIC,"Wackamole_init");
 
   /* Set defaults */
   Maturity_timeout.sec = 5*60;
@@ -986,6 +1067,10 @@ static	void    Wackamole_init()
 
   spread_lock = 0;
 
+#ifdef USE_EMBEDDED_PERL
+  perl_startup();
+#endif
+
   Get_conf(File_name, &My);
 
   /* start up control socket */
@@ -1016,8 +1101,10 @@ static  void    Usage(int argc, char *argv[])
     {
       argv++;
 
-      if( !strncmp( *argv, "-u", 2 ) )
-	{
+      if( !strncmp( *argv, "-v", 2 ) ) {
+          notice();
+          exit(0);
+        }else if( !strncmp( *argv, "-u", 2 ) ) {
 	  strcpy( User, argv[1] );
 	  argc--; argv++;
 	}else if( !strncmp( *argv, "-s", 2 ) ){
@@ -1036,10 +1123,13 @@ static  void    Usage(int argc, char *argv[])
 	}else if(!strncmp(*argv, "-d", 2) ){
 	  Debug = 1;
 	}else{
-	  Alarm(EXIT, "Usage: wackamole\n%s\n%s\n%s",
-		  "\t[-u <private name>]  : unique (in this machine) private name",
-		  "\t[-s <address>]    : either port or port@machine"
-		  "\t[-c <confuguration file name>]   :file used instread of wackamole.conf");
+	  wack_alarm(PRINT, "Usage: wackamole\n"
+		  "\t-c <filename>     : file used instread of wackamole.conf\n"
+		  "\t-d                : debug mode and don't daemonize\n"
+		  "\t-s <address>      : either port or port@machine\n"
+                  "\t-u <username>     : connect to Spread as user username\n"
+                  "\t-v                : version information\n");
+          exit(0);
 	}
     }
 }
@@ -1053,15 +1143,15 @@ static  void    Print_alloc()
   address    Addr;
   char       IP[16];
 
-  Alarm(LOGIC, "Current allocation table:" );
+  wack_alarm(LOGIC, "Current allocation table:" );
   for ( i = 0; i < Num_pseudo; i++ )
     {
       Addr = Allocation_table[i].pseudo_address;
       String_addr( Addr, IP );
-      Alarm(LOGIC, "Pseudo address: %s", IP );
+      wack_alarm(LOGIC, "Pseudo address: %s", IP );
       Addr = Allocation_table[i].real_address;
       String_addr( Addr, IP );
-      Alarm(LOGIC, "\tis allocated to %s with priority %d.", 
+      wack_alarm(LOGIC, "\tis allocated to %s with priority %d.", 
 	      IP, Allocation_table[i].claim_priority );
     }
 #endif
@@ -1084,20 +1174,23 @@ static  void    Print_alloc()
 /*************************************************/
 
 static void Sig_handler(int signum) {
-  Alarm(DEBUG,"Sig_handler called");
+  wack_alarm(WACK_DEBUG,"Sig_handler called");
   if(signum == SIGINT)
-    Alarm(DEBUG, "SIGINT Detected!");
+    wack_alarm(WACK_DEBUG, "SIGINT Detected!");
   if(signum == SIGTERM)
-    Alarm(DEBUG,"SIGTERM Detected!");  
+    wack_alarm(WACK_DEBUG,"SIGTERM Detected!");  
   if(signum == SIGBUS)
-    Alarm(DEBUG,"SIGBUS Detected!");
+    wack_alarm(WACK_DEBUG,"SIGBUS Detected!");
   if(signum == SIGQUIT) 
-    Alarm(DEBUG,"SIGQUIT Detected!");  
+    wack_alarm(WACK_DEBUG,"SIGQUIT Detected!");  
   if(signum == SIGSEGV) 
-    Alarm(DEBUG,"SIGSEGV Detected!");
+    wack_alarm(WACK_DEBUG,"SIGSEGV Detected!");
 
   Clean_up();
   global_exit = 1;
+#ifdef USE_EMBEDDED_PERL
+  perl_shutdown();
+#endif
   exit(0);
 }
 
@@ -1107,7 +1200,7 @@ static void Sig_handler(int signum) {
 void Clean_up(){
   
   int cl_index;
-  Alarm(DEBUG, "Clean_up called");
+  wack_alarm(WACK_DEBUG, "Clean_up called");
  
   for(cl_index = 0; cl_index < Num_pseudo; cl_index++){
     cancel_spoofer(&Old_table[cl_index]);
@@ -1128,9 +1221,9 @@ void    handle_reconnect(int a, void *d) {
 void    Spread_reconnect(int ret){
 
   if(ret != -8 && ret != -11 && ret != -2)
-    Alarm(EXIT, "Spread_reconnnect: Unexpected Error (%d)", ret);
+    wack_alarm(EXIT, "Spread_reconnnect: Unexpected Error (%d)", ret);
   
-  Alarm(PRINT,"connecting to %s", Spread_name);
+  wack_alarm(PRINT,"connecting to %s", Spread_name);
     
   Clean_up();
 
@@ -1149,11 +1242,11 @@ void    Spread_reconnect(int ret){
       SP_error( ret );
       SP_disconnect( Mbox );
       Mbox = -1;
-      Alarm(PRINT, "Spread join on reconnect failed [%d].", ret);
+      wack_alarm(PRINT, "Spread join on reconnect failed [%d].", ret);
       return handle_reconnect(0, NULL);
     }
   } else {
-    Alarm(PRINT, "Spread connect failed [%d].", ret);
+    wack_alarm(PRINT, "Spread connect failed [%d].", ret);
     return handle_reconnect(0, NULL);
   }
 
@@ -1174,7 +1267,7 @@ void    Spread_reconnect(int ret){
 /********************************************************/
 
 void Bye(){
-  Alarm(DEBUG, "Disconnecting from Spread");
+  wack_alarm(WACK_DEBUG, "Disconnecting from Spread");
   SP_disconnect(Mbox);
 }
 
